#!/usr/bin/env python
"""
Copyright (c) 2017-2021 HoloViz developers.

Copyright (c) 2013 Nathan Goldbaum. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

   * Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above
copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the
distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from __future__ import annotations

import copy
import glob
import io
import json
import os
import re
import shutil
import string
import sys
import typing

from contextlib import contextmanager

import nbformat

from docutils.parsers.rst import Directive, directives
from docutils.statemachine import string2lines
from docutils.utils import new_document
from myst_nb.sphinx_ import Parser
from nbconvert import NotebookExporter, PythonExporter
from nbconvert.preprocessors import (
    CellExecutionError, ExecutePreprocessor, Preprocessor,
)
from sphinx.util import logging

from .cmd import _prepare_paths, hosts

if typing.TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import Tuple

logger = logging.getLogger(__name__)

NOTEBOOK_VERSION = 4

interactivity_warning_binder = (
    'This web page was generated from a Jupyter notebook and not all '
    'interactivity will work on this website. <a href="{download_link}">'
    'Right-click to download and run</a> or <a href="{binder_link}">'
    'Launch on Binder</a> for full Python-backed interactivity.'
)

interactivity_warning = (
    'This web page was generated from a Jupyter notebook and not all '
    'interactivity will work on this website. <a href="{download_link}">'
    'Right click to download and run locally</a> for full Python-backed '
    'interactivity.'
)

#from nbformat.v4 import output_from_msg
class ExecutePreprocessor1000(ExecutePreprocessor):
    """Sigh"""
    _ipython_startup = None

    @property
    def kc(self):
        return self._kc

    @kc.deleter
    def kc(self):
        del self._kc

    @kc.setter
    def kc(self, v):
        self._kc = v
        if v is not None and self._ipython_startup is not None:
            msg_id = self._kc.execute( # noqa: a mess
                self._ipython_startup,silent=False,store_history=False,allow_stdin=False,stop_on_error=True)


    def handle_comm_msg(self, outs, msg, cell_index):
        """
        Comm messages are not handled correctly in some cases so we
        just ignore them during export.
        """
        pass


class SkipOutput(Preprocessor):
    """A transformer to skip the output for cells containing a certain string"""

    def __init__(self, substring=None, **kwargs):
        self.substring = substring
        super(SkipOutput, self).__init__(**kwargs)

    def preprocess_cell(self, cell, resources, index):
        if cell['cell_type'] == 'code':
            if self.substring in cell['source']:
                cell['outputs'] = []
        return cell, resources

    def __call__(self, nb, resources):
        return self.preprocess(nb,resources)


class NotebookSlice(Preprocessor):
    """A transformer to select a slice of the cells of a notebook"""


    def __init__(self, substring=None, end=None, offset=0, **kwargs):
        self.substring = substring
        self.end = end
        self.offset = offset
        super(NotebookSlice, self).__init__(**kwargs)

    def _find_slice(self, nbc, substring, endstr):
        start = 0 if substring is None else None
        end = None
        try:
            end = int(endstr)
            count = 0
        except Exception:
            count = None

        for ind, cell in enumerate(nbc.cells):
            source = cell['source']
            if start is None and substring in source:
                start = ind
            if None not in [start, count]:
                if end is not None and count >= end:
                    end = ind
                    break
                count += 1
            elif start is not None:
                if endstr is None:
                    break
                elif endstr in source:
                    end = ind + 1
                    break

        if start is None:
            raise Exception('Invalid notebook slice start')
        if end is None and endstr is not None:
            raise Exception('Invalid notebook slice end string match')

        return (max([start,self.offset]),end)

    def preprocess(self, nb, resources):
        nbc = copy.deepcopy(nb)
        start,end = self._find_slice(nbc, self.substring, self.end)
        nbc.cells = nbc.cells[start:end]
        return nbc, resources

    def __call__(self, nb, resources):
        return self.preprocess(nb,resources)


class FixNotebookLinks(Preprocessor):
    """
    Fixes relative notebook links by pointing to ReST or Markdown
    source files and if necessary stripping leading numbering
    (e.g. 0_notebook.ipynb) which are stripped by the generate-rst
    script.
    """

    file_types = ['rst', 'md']

    # Regex for matching markdown notebook links.
    # Example input "[sometext](../../somepath/subject.ipynb#spam)""
    nb_regex = re.compile((
                r"(\[.+?\]\()"   # "[sometext]("
                r"(.+?\.ipynb)"  # "../../somepath/subject.ipynb"
                r"(#*?.*?\))"    # "#spam)"
    ))

    def __init__(self, nb_path: str, **kwargs):
        # nb_path: the directory of the notebook, without a trailing slash
        self.nb_path = nb_path
        super(FixNotebookLinks, self).__init__(**kwargs)

    def preprocess_cell(self, cell, resources, index):
        if cell['cell_type'] != 'markdown':
            return cell, resources

        cell['source'] = self.replace_notebook_links(
            markdown_text=cell['source'],
            rootdir=self.nb_path
        )

        return cell, resources

    @classmethod
    def replace_notebook_links(cls, markdown_text: str, rootdir:str) -> str:
        """Replaces notebook links in a markdown text.

        Parameters
        ----------
        markdown_text: str
            Markdown text from a notebook markdown cell.
        rootdir:
            The absolute path to the directory where the notebook that is being
            processed (which contains links) is located at, without a trailing
            slash.
        """
        for nb_link, nb_filepath in cls._extract_links(markdown_text):

            target_relpath = cls._get_sourcefile(rootdir, nb_filepath)
            if not target_relpath:
                logger.warn('Source file for "%s" not found (path relative to "%s")', nb_filepath, rootdir)
                continue

            new_link = cls._create_target_link(nb_link, target_relpath)
            markdown_text = markdown_text.replace(nb_link, new_link)

        return markdown_text

    @classmethod
    def _get_sourcefile(cls, rootdir: str, nb_filepath: str) -> str | None:
        """"Get the source (.rst / .md) file path for a notebook.

        Parameters
        ----------
        rootdir:
            The location of the root notebook (where the link markdown is at)
        nb_filepath:
            Location of the notebook which source to find, relative to rootdir.

        Returns
        -------
        source_relpath: str | None
            The source file path relative to root notebook. If the source file
            is not found, issues a warning and returns none.
        """
        for source_relpath in cls._iter_source_file_candidates(nb_filepath):
            target_abspath = os.path.normpath(os.path.join(rootdir, source_relpath))
            if cls._file_exists(target_abspath):
                return source_relpath

    @classmethod
    def _extract_links(cls, markdown_text: str) -> Iterable[Tuple[str, str]]:
        """Extract links to Notebook files (.ipynb) from markdown text and
        yield them. Returns the full markdown link and the link target
        separately.

        Examples
        --------
        Markdown link: "[a](../foo/b.ipynb#spam)"
        Link target: "..foo/b.ipynb"

        String "foo [a](b.ipynb) bar [c](..foo/d.ipynb#spam)" would return
        iterable, which when converted to a list would be: [
             [('[a](b.ipynb)', 'b.ipynb')]
             [('[c](../foo/d.ipynb#spam)', '../foo/d.ipynb')]
        ]
        """

        for match in cls.nb_regex.finditer(markdown_text):
            markdown_link = ''.join(match.groups())
            link_target = match.group(2)
            yield markdown_link, link_target

    @classmethod
    def _iter_source_file_candidates(cls, nb_filepath: str) -> Iterable[str]:
        """Gets potential link targets corresponding to a notebook file path.
        Potential link targets (source files paths) are formed by using the
        relative path of the link target notebook, and adding the file
        extensions from `cls.file_types`. In addition if the the filename part
        of the `nb_filepath` starts with digits + one of ('-', '_', ' '),
        the stem without the number prefix is used to form a potential link
        target.

        Parameters
        ----------
        nb_filepath:
            The filepath of a link target a notebook. Must end with ".ipynb".
            The links are relative to the directory of the notebook which
            contains the markdown link. Example: "../foo/10-Some.ipynb"

        Example
        -------
        >>> list(_iter_source_file_candidates("../foo/01-Some_Notebook.ipynb"))
        [
            "../foo/01-Some_Notebook.rst",
            "../foo/01-Some_Notebook.md",
            "../foo/Some_Notebook.rst",
            "../foo/Some_Notebook.md",
        ],
        """

        if not nb_filepath.endswith('.ipynb'):
            raise ValueError('nb_filepath must end with .ipynb')

        nb_path_without_extension = nb_filepath[:-6]
        for extension in cls.file_types:
            yield f"{nb_path_without_extension}.{extension}"

        for extension in cls.file_types:
            directory, stem = os.path.split(nb_path_without_extension)
            match = re.match(r"\d+[ -_](.*)", stem)
            if match:
                yield os.path.join(directory, match.group(1) + '.' + extension)


    @classmethod
    def _create_target_link(cls, nb_link: str, target_relpath:str) -> str:
        """Using a markdown link `nb_link`, create a markdown link which
        changes the link target file to be `target_relpath`.

        Example
        -------
        If nb_link is "[sometext](../../somepath/01-subject.ipynb#spam)", and
        the target_filename is "../otherpath/subject.rst", returns
        "[sometext](../otherpath/subject.rst#spam)"

        Notes:
        * If you pass as nb_link text which contains no markdown links, a
          ValueError will be raised
        * If you pass as nb_link text which contains more than one markdown
          links, the behaviour is undefined.
        """

        match = cls.nb_regex.match(nb_link)

        if not match:
            raise ValueError(f'Not a valid markdown link!: {nb_link}')

        return f"{match.group(1)}{target_relpath}{match.group(3)}"

    @staticmethod
    def _file_exists(file_path: str) -> bool:
        # Makes faking files in tests easier.
        return os.path.isfile(file_path)

    def __call__(self, nb, resources):
        return self.preprocess(nb,resources)


class FixBackticksInDetails(Preprocessor):
    """A preprocessor to make backticks in details cells work."""

    def comment_out_details(self, source):
        """
        Given the source of a cell, comment out  any lines that contain <details>
        """
        filtered = []
        for line in source.splitlines():
            if "details>" in line:
                line ='<!--' + line + '-->'
            filtered.append(line)
        return '\n'.join(filtered)

    def preprocess_cell(self, cell, resources, index):
        if cell['cell_type'] == 'markdown':
            if '<details>' in cell['source'] and "```" in cell['source']:
                cell['source'] = self.comment_out_details(cell['source'])
        return cell, resources

    def __call__(self, nb, resources):
        return self.preprocess(nb,resources)


def get_download_link(relpath, org, branch, repo, examples, host):
    info = (hosts[host], org, repo, branch, examples, relpath)
    return '%s/%s/%s/%s/%s/%s' % info


def get_binder_link(relpath, org, repo, branch, examples):
    # SVG logo: https://mybinder.org/badge_logo.svg
    url="https://mybinder.org/v2/gh/{org}/{repo}/{branch}?filepath={examples}/{relpath}"
    return  url.format(org=org,repo=repo,branch=branch,relpath=relpath,examples=examples)


def formatted_link(path):
    return "`%s <%s>`__" % (os.path.basename(path), path)


def nb_to_python(nb_path):
    """convert notebook to python script"""
    exporter = PythonExporter()
    output, resources = exporter.from_filename(nb_path)
    return output


def evaluate_notebook(nb_path, dest_path=None, skip_exceptions=False,
                      skip_execute=None, timeout=300, ipython_startup=None,
                      patterns_to_take_with_me=None):

    if patterns_to_take_with_me is None:
        patterns_to_take_with_me = []

    notebook = nbformat.read(nb_path, as_version=4)
    kwargs = dict(timeout=timeout,
                  kernel_name='python%s'%sys.version_info[0],
                  allow_errors=skip_exceptions)

    cwd = os.getcwd()
    filedir, filename = os.path.split(nb_path)
    os.chdir(filedir)
    not_nb_runner = ExecutePreprocessor1000(**kwargs)
    if ipython_startup is not None:
        not_nb_runner._ipython_startup = ipython_startup

    if not os.path.isfile(dest_path):
        print('INFO: Writing evaluated notebook to {dest_path!s}'.format(
            dest_path=os.path.abspath(dest_path)))
        try:
            if not skip_execute:
                not_nb_runner.preprocess(notebook,{})
        except CellExecutionError as e:
            print('')
            print(e)
        os.chdir(cwd)

        if skip_execute:
            with open(dest_path,'w', encoding='utf-8') as f:
                nbformat.write(notebook, f)
        else:
            ne = NotebookExporter()
            newnb, _ = ne.from_notebook_node(notebook)
            with open(dest_path, 'w', encoding='utf-8') as f:
                f.write(newnb)
            for pattern in patterns_to_take_with_me:
                for f in glob.glob(os.path.join(os.path.dirname(nb_path),pattern)):
                    print("mv %s %s"%(f, os.path.dirname(dest_path)))
                    shutil.move(f,os.path.dirname(dest_path))
    else:
        print('INFO: Skipping existing evaluated notebook {dest_path!s}'.format(
            dest_path=os.path.abspath(dest_path)))


@contextmanager
def disable_execution(env):
    # Just to make sure that the notebook, which should already be executed
    # by ExecutePreprocessor1000 isn't re-executed again. It should not
    # happen by default if myst-nb's execution_mode isn't touched as not
    # reexecuting notebooks fully executed is its default behavior.
    if hasattr(env, 'mystnb_config'):
        old_exec_mode = env.mystnb_config.execution_mode
        env.mystnb_config.execution_mode = 'off'
    try:
        yield
    finally:
        # Restore execution mode
        if hasattr(env, 'mystnb_config'):
            env.mystnb_config.execution_mode = old_exec_mode

@contextmanager
def patch_project_doc2path(env, nb_path):
    # Ugly patch required as myst-nb obtains a reader that
    # is inferred by sphinx from the document. Sphinx literally
    # loops through a glob() result looking for files in a particular
    # order, starting from .rst. As the NotebookDirective is embedded
    # in a .rst, it finds that file instead of the Notebook to be parsed.
    # Overriding the doc2path Project method to directly return the
    # absolute notebook path.
    old_doc2path = env.project.doc2path

    def new_doc2path(docname: str, absolute: bool):
        return nb_path

    env.project.doc2path = new_doc2path
    try:
        yield
    finally:
        env.project.doc2path = old_doc2path


def render_notebook(nb_path, document, preprocessors=[]):
    env = document.settings.env
    doc = new_document(nb_path, document.settings)

    # Load notebook and run preprocessors
    with open(nb_path, encoding='utf-8') as f:
        text = f.read()

    ntbk = nbformat.reads(text, as_version=NOTEBOOK_VERSION)
    for preprocessor in preprocessors:
        ntbk, _ = preprocessor(ntbk, {})
    sio = io.StringIO(json.dumps(ntbk))

    parser = Parser()
    parser.env = env

    with disable_execution(env), patch_project_doc2path(env, nb_path):
        parser.parse(sio.read(), doc)

    return doc.children


class NotebookDirective(Directive):
    """Insert an evaluated notebook into a document

    This uses runipy and nbconvert to transform a path to an unevaluated notebook
    into html suitable for embedding in a Sphinx document.
    """
    required_arguments = 2
    optional_arguments = 6
    option_spec = {
        'skip_exceptions' : directives.flag,
        'substring': str,
        'end': str,
        'skip_execute': bool,
        'skip_output': str,
        'offset': int,
        'disable_interactivity_warning': bool
    }

    def link_rst(self, nb_basename, nb_abs_path, dest_path):
        # Process file inclusion options
        include_opts = self.arguments[2:]
        include_nb = True if 'ipynb' in include_opts else False
        include_script = True if 'py' in include_opts else False

        link_rst = ''
        if len(include_opts):
            link_rst = 'Direct Downloads: ('

        if include_nb:
            link_rst += formatted_link(nb_basename) + '; '

        if include_script:
            dest_path_script = string.replace(dest_path, '.ipynb', '.py')
            rel_path_script = string.replace(nb_basename, '.ipynb', '.py')
            script_text = nb_to_python(nb_abs_path)
            with open(dest_path_script, 'w', encoding='utf-8') as f:
                f.write(script_text.encode('utf-8'))

            link_rst += formatted_link(rel_path_script)

        if len(include_opts):
            link_rst = ')'

        return link_rst

    def preprocessors(self, dest_dir):
        preprocessors = [FixBackticksInDetails(), FixNotebookLinks(dest_dir)]
        if self.options.get('substring') or self.options.get('offset'):
            preprocessors.append(
                NotebookSlice(
                    self.options.get('substring'),
                    self.options.get('end'),
                    self.options.get('offset')
                )
            )
        if self.options.get('skip_output'):
            preprocessors.append(SkipOutput(self.options['skip_output']))
        return preprocessors

    def interactivity_warning(self, nb_abs_path):
        project_name = os.environ.get('PROJECT_NAME','')
        project_root = os.environ.get('PROJECT_ROOT','')
        host = os.environ.get('HOST','GitHub')
        branch = os.environ.get('BRANCH','main')
        repo = os.environ.get('REPO','')
        org = os.environ.get('ORG','')
        doc = os.environ.get('DOC','doc')
        examples = os.environ.get('EXAMPLES','examples')
        examples_assets = os.environ.get('EXAMPLES_ASSETS','assets')
        binder = os.environ.get('BINDER','none')

        if repo == '' or project_name == '':
            project_name = repo = project_name or repo
        if org == '':
            org = repo

        paths = _prepare_paths(project_root, examples=examples,
                               doc=doc, examples_assets=examples_assets)
        relpath = os.path.relpath(nb_abs_path, paths['examples'])
        dl_link = get_download_link(relpath, org, branch, repo, examples, host)
        binder_link = get_binder_link(relpath, org, repo, branch, examples)

        if binder == 'none':
            inner_msg = interactivity_warning.format(download_link=dl_link)
        else:
            inner_msg = interactivity_warning_binder.format(download_link=dl_link,
                                                            binder_link=binder_link)
        inner_msg = inner_msg.replace('\n', '')
        scroller = f'<div id="scroller-right">{inner_msg}</div>'
        return f'\n.. raw:: html\n\n    {scroller}'

    def run(self):
        # Process paths and directories
        rst_file = os.path.abspath(self.state.document.current_source)
        rst_dir = os.path.dirname(rst_file)
        nb_abs_path = os.path.abspath(os.path.join(rst_dir, self.arguments[1]))
        nb_filepath, nb_basename = os.path.split(nb_abs_path)

        dest_dir = rst_dir
        dest_path = os.path.join(dest_dir, nb_basename)

        os.makedirs(dest_dir, exist_ok=True)

        # Evaluate Notebook and insert into Sphinx doc
        evaluate_notebook(
            nb_abs_path, dest_path,
            skip_exceptions='skip_exceptions' in self.options,
            skip_execute=self.options.get('skip_execute'),
            timeout=setup.config.nbbuild_cell_timeout,
            ipython_startup=setup.config.nbbuild_ipython_startup,
            patterns_to_take_with_me=setup.config.nbbuild_patterns_to_take_along
        )

        preprocessors = self.preprocessors(dest_dir)
        rendered_nodes = render_notebook(
            dest_path, self.state.document, preprocessors
        )

        link_rst = self.link_rst(nb_basename, nb_abs_path, dest_path)
        if 'disable_interactivity_warning' not in self.options:
            link_rst += self.interactivity_warning(nb_abs_path)

        # Insert evaluated notebook HTML into Sphinx
        if link_rst:
            include_lines = string2lines(link_rst, convert_whitespace=True)
            self.state_machine.insert_input(include_lines, rst_file)

        # add dependency
        self.state.document.settings.record_dependencies.add(nb_abs_path)

        # TODO: doubt this is doing anything
        # clean up png files left behind by notebooks.
        png_files = glob.glob("*.png")
        for file in png_files:
            os.remove(file)

        return rendered_nodes


def setup(app):
    setup.app = app
    setup.config = app.config
    setup.confdir = app.confdir

    app.add_config_value('nbbuild_cell_timeout',300,'html')
    app.add_config_value('nbbuild_ipython_startup',"from nbsite.ipystartup import *",'html')
    app.add_config_value('nbbuild_patterns_to_take_along',["*.json", "json_*"],'html')

    app.add_directive('notebook', NotebookDirective)



# -*- coding: utf-8 -*-

import datetime
import importlib
import inspect
import os
import pathlib as _pathlib
import subprocess
import sys

import nbsite as _nbsite

from nbsite import nbbuild
from nbsite.util import base_version  # noqa

NBSITE_DIR = _pathlib.Path(_nbsite.__file__).parent


def holoviz_icon_white(cur_file):
    """Return the path as a string of the holoviz white icon SVG."""
    doc_root = _pathlib.Path(cur_file).parent
    shared_static_dir = _pathlib.Path(os.path.relpath(NBSITE_DIR / '_shared_static', start=doc_root))
    icon_path = shared_static_dir / 'holoviz-icon-white.svg'
    return str(icon_path)


def setup(app):
    try:
        from nbsite.paramdoc import param_formatter, param_skip
        app.connect('autodoc-process-docstring', param_formatter)
        app.connect('autodoc-skip-member', param_skip)
    except ImportError:
        print('no param_formatter (no param?)')

    nbbuild.setup(app)
    app.connect("builder-inited", remove_mystnb_static)

def remove_mystnb_static(app):
    # Ensure our myst_nb.css is loaded by removing myst_nb static_path
    # from config
    app.config.html_static_path = [
        p for p in app.config.html_static_path if 'myst_nb' not in p
    ]

extensions = [
    'myst_nb',
    'sphinx_design',
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.linkcode',
    'sphinx.ext.inheritance_diagram',
    'sphinx_copybutton',
    'sphinxext.rediraffe',
]

# Default theme is the PyData Sphinx Theme
html_theme = "pydata_sphinx_theme"

inheritance_graph_attrs = dict(
    rankdir="LR",
    size='"12.0, 12.0"',
    fontsize=18
)

default_edge_attrs = {
    'arrowsize': 1.0,
    'style': '"setlinewidth(0.5)"',
}

source_suffix = '.rst'

master_doc = 'index'

pygments_style = 'sphinx'

exclude_patterns = ['_build']

html_static_path = [
    os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '_shared_static'
        )
    )
]

templates_path = [
    os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            '_shared_templates'
        )
    )
]

html_context = {}

html_css_files = [
    'nbsite.css',
    'notebook.css',
    'gallery.css',
    'alert.css',
    'dataframe.css',
    'scroller.css'
]

# A single line footer that includes the copyright and the last updated date.
html_theme_options = {
    "footer_start": [
        "copyright-last-updated",
    ],
    # To avoid warning as this is the new default
    # See https://github.com/pydata/pydata-sphinx-theme/issues/1492
    "navigation_with_keys": False,
}

# Default is "**": ["sidebar-nav-bs", "sidebar-ethical-ads"]
# Overriding the index sidebar to include the toctree on the landing page.
# The layout.html template in pydata-sphinx-theme removes the default
# sidebar-nav-bs.html template.
html_sidebars = {
    "index": ["sidebar-nav-bs-alt"],
    "**": ["sidebar-nav-bs-alt"],
}

# To be reused in a conf.py file to define the `copyright` string reused
# by sphinx to populate the footer content:
# copyright_years['start_year'] = '2000'
# copyright = copyright_fmt.format(**copyright_years)
copyright_years = {'current_year': str(datetime.date.today().year)}
copyright_fmt = "{start_year}-{current_year} Holoviz contributors"

# Format of the last updated date in the footer.
html_last_updated_fmt = '%Y-%m-%d'

rediraffe_redirects = {}

suppress_warnings = [
    # Ignore: (WARNING/2) Document headings start at H2, not H1
    "myst.header",
    # Ignores: skipping unknown output mime type: application/vnd.holoviews_exec.v0+json
    "mystnb.unknown_mime_type"
]

GIT_ROOT_CMD = "git rev-parse --show-toplevel"
GITHUB_BASE_URL = "https://github.com/holoviz/"

def get_module_object(modname, fullname):
    """Retrieve the Python object based on its module and fullname."""
    submod = sys.modules.get(modname)
    if submod is None:
        return None
    obj = submod
    for part in fullname.split("."):
        obj = getattr(obj, part, None)
        if obj is None:
            return None
    return obj

def get_source_info(obj):
    """Get the source file and line numbers for a Python object."""
    try:
        fn = inspect.getsourcefile(inspect.unwrap(obj))
        source, lineno = inspect.getsourcelines(obj)
    except (TypeError, OSError):
        return None, None, None
    return fn, source, lineno

def get_repo_dir():
    """Get the root directory of the current Git repository."""
    try:
        output = subprocess.check_output(GIT_ROOT_CMD.split(" ")).strip().decode('utf-8')
        return _pathlib.Path(output).name
    except subprocess.CalledProcessError:
        raise RuntimeError("Unable to determine the repository directory")

def linkcode_resolve(domain, info):
    """Resolve the link to the source code for a given Python object in documentation."""
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    obj = get_module_object(modname, fullname)
    if obj is None:
        return None

    fn, source, lineno = get_source_info(obj)
    if not fn:
        return None

    linespec = f"#L{lineno}-L{lineno + len(source) - 1}" if lineno else ""

    package = get_repo_dir()

    ppath = importlib.__import__(package).__file__
    pver = importlib.__import__(package).__version__

    fn = os.path.relpath(fn, start=os.path.dirname(ppath))

    if "+" in pver:
        return f"{GITHUB_BASE_URL}{package}/blob/main/{package}/{fn}{linespec}"
    else:
        return f"{GITHUB_BASE_URL}{package}/blob/v{pver}/{package}/{fn}{linespec}"



import asyncio
import hashlib
import io
import json
import os
import pathlib
import sys
import traceback as tb
import warnings

from collections import defaultdict
from html import escape
from multiprocessing import Pipe, get_context
from pathlib import Path
from typing import (
    Any, Dict, List, Tuple,
)

import param
import portalocker

from bokeh.document import Document
from bokeh.embed.bundle import _bundle_extensions
from bokeh.embed.util import standalone_docs_json_and_render_items
from bokeh.model import Model
from docutils import nodes
from docutils.parsers.rst import Directive, roles
from jinja2.environment import Environment
from jinja2.loaders import FileSystemLoader
from packaging.version import Version
from panel.config import config, panel_extension as extension
from panel.io.convert import BOKEH_VERSION
from panel.io.mime_render import exec_with_return, format_mime
from panel.io.resources import CDN_DIST, Resources, set_resource_mode
from panel.pane import HoloViews, Interactive, panel as as_panel
from panel.reactive import ReactiveHTML
from panel.viewable import Viewable, Viewer
from sphinx.application import Sphinx

HERE = Path(__file__).parent

CONVERT_PANE_TYPES = (HoloViews, Interactive)

try:
    from panel.param import ReactiveExpr
    CONVERT_PANE_TYPES += (ReactiveExpr,)
except Exception:
    pass

def get_env() -> Environment:
    ''' Get the correct Jinja2 Environment, also for frozen scripts.
    '''
    return Environment(loader=FileSystemLoader(HERE))

_env = get_env()
_env.trim_blocks = True
_env.lstrip_blocks = True

WEB_WORKER_TEMPLATE = _env.get_template('WebWorker.js')
WORKER_HANDLER_TEMPLATE = _env.get_template('WorkerHandler.js')
SERVICE_WORKER_TEMPLATE = _env.get_template('ServiceWorker.js')
SERVICE_HANDLER_TEMPLATE = _env.get_template('ServiceHandler.js')
WEB_MANIFEST_TEMPLATE = _env.get_template('site.webmanifest')

JS_MODULE_TAG = """
<script type="module" src="{file}"></script>"""
JS_MODULE_EXPORT = """
<script type="module">
  import {name} from "{file}";
  window.{name} = {name}
</script>
"""

RESOURCE_FILE = '.extra_resources.json'

bokeh_version = Version(BOKEH_VERSION)
if bokeh_version.is_devrelease or bokeh_version.is_prerelease:
    bk_prefix = 'dev'
else:
    bk_prefix = 'release'

DEFAULT_PYODIDE_CONF = {
    'PYODIDE_URL': 'https://cdn.jsdelivr.net/pyodide/v0.25.0/full/pyodide.js',
    'autodetect_deps': True,
    'enable_pwa': True,
    'requirements': ['panel', 'pandas'],
    'precache': [],
    'scripts': [
        f'https://cdn.bokeh.org/bokeh/{bk_prefix}/bokeh-{BOKEH_VERSION}.min.js',
        f'https://cdn.bokeh.org/bokeh/{bk_prefix}/bokeh-widgets-{BOKEH_VERSION}.min.js',
        f'https://cdn.bokeh.org/bokeh/{bk_prefix}/bokeh-tables-{BOKEH_VERSION}.min.js',
        f'{CDN_DIST}panel.min.js'
    ],
    'extra_css': [],
    'cache_patterns': [
        CDN_DIST,
        'https://cdn.bokeh.org/bokeh/',
        'https://cdn.jsdelivr.net/pyodide/',
        'https://files.pythonhosted.org/packages/',
        'https://pypi.org/pypi/'
    ],
    'preamble': "",
    'setup_code': "",
    'warn_message': "Executing this cell will download Python runtime (typically 40+ MB).",
    'requires': {}
}

def extract_extensions(code: str) -> List[str]:
    """
    Extracts JS and CSS dependencies of Panel extensions from code snippets
    containing pn.extension calls.
    """
    js, js_exports, js_modules, css = [], {}, {}, []
    with set_resource_mode('cdn'):
        for name, model in Model.model_class_reverse_map.items():
            if hasattr(model, '__javascript__'):
                js += model.__javascript__
            if hasattr(model, '__css__'):
                css += model.__css__
            if hasattr(model, '__javascript_modules__'):
                js_modules.update({
                    jsm.split("/")[-1][:-3]: jsm for jsm in model.__javascript_modules__
                    if jsm.endswith('.js')
                })
            if hasattr(model, '__javascript_module_exports__'):
                js_exports.update(
                    dict(zip(model.__javascript_module_exports__,
                             model.__javascript_modules__))
                )
        for model in param.concrete_descendents(ReactiveHTML).values():
            if not model._loaded():
                continue
            js += getattr(model, '__javascript__', []) or []
            css += getattr(model, '__css__', []) or []
    if config.design and hasattr(config.design, 'resolve_resources'):
        design_resources = config.design().resolve_resources(
            cdn=True, include_theme=True
        )
        js += list(design_resources['js'].values())
        css += list(design_resources['css'].values())
        js_modules.update(design_resources['js_modules'])
    resources = Resources(mode='cdn')
    extensions = _bundle_extensions(None, resources)
    js += [cdn_url for bundle in extensions if bundle.cdn_url and
           '@holoviz/panel@' not in (cdn_url:= str(bundle.cdn_url))]
    global_exports = [
        extension._globals[ext][0] for ext, imp in extension._imports.items()
        if imp in sys.modules and ext in extension._globals
    ]
    return js, js_exports, js_modules, css, global_exports

def _model_json(model: Model, target: str) -> Tuple[Document, str]:
    """
    Renders a Bokeh Model to JSON representation given a particular
    DOM target and returns the Document and the serialized JSON string.

    Arguments
    ---------
    model: bokeh.model.Model
        The bokeh model to render.
    target: str
        The id of the DOM node to render to.

    Returns
    -------
    document: Document
        The bokeh Document containing the rendered Bokeh Model.
    model_json: str
        The serialized JSON representation of the Bokeh Model.
    """
    doc = Document()
    model.server_doc(doc=doc)
    model = doc.roots[0]
    docs_json, _ = standalone_docs_json_and_render_items(
        [model], suppress_callback_warning=True
    )
    doc_json = list(docs_json.values())[0]

    return doc, json.dumps(dict(
        target_id = target,
        root_id   = model.ref['id'],
        doc       = doc_json,
        version   = BOKEH_VERSION,
    ))

def write_resources(out_dir, source, resources):
    """
    Writes resources extracted from process to a shared JSON file to
    which will be read on document write. On windows uses in-memory
    variable since parallelization is not supported.

    Arguments
    ---------
    out_dir: str
        The build directory
    source: str
        The source file the resources are assosicated with.
    resources: dict[str, any]
        The resources to add to the resource dictionary.
    """
    out_path = pathlib.Path(os.fspath(out_dir))
    out_path.mkdir(exist_ok=True)
    resources_file = out_path / RESOURCE_FILE
    existing = resources_file.is_file()
    with portalocker.Lock(resources_file, 'a+') as rfile:
        rfile.seek(0)

        # Load existing resources from file
        all_resources = {}
        if existing:
            all_resources = json.load(rfile)
            rfile.seek(0)
            rfile.truncate()

        if source in all_resources:
            # Merge with existing resources for source file
            source_resources = all_resources[source]
            source_resources['css'] += [css for css in resources['css'] if css not in source_resources['css']]
            source_resources['js'] += [js for js in resources['js'] if js not in source_resources['js']]
            source_resources['js_exports'].update(resources['js_exports'])
            source_resources['js_modules'].update(resources['js_modules'])
        else:
            # Add new resources for this source file
            all_resources[source] = source_resources = resources
        json.dump(all_resources, rfile)
        rfile.flush()
        os.fsync(rfile.fileno())



def _option_boolean(arg):
    if not arg or not arg.strip():
        # no argument given, assume used as a flag
        return True
    elif arg.strip().lower() in ("no", "0", "false"):
        return False
    elif arg.strip().lower() in ("yes", "1", "true"):
        return True
    else:
        raise ValueError('"%s" unknown boolean' % arg)

class PyodideDirective(Directive):

    has_content = True

    option_spec = {
        'skip-embed': _option_boolean
    }

    _exec_state = defaultdict(lambda: {
        'source': None,
        'context': {},
        'count': 0,
        'total': None,
        'process': None,
        'conn': None,
        'cache': {},
        'hash': None
    })

    @classmethod
    def _execution_process(cls, pipe):
        """
        Process execution loop to run in separate process that is used
        to evaluate code.
        """
        async def loop():
            while True:
                msg = pipe.recv()
                if msg['type'] == 'close':
                    break
                elif msg['type'] != 'execute':
                    continue
                stdout = io.StringIO()
                stderr = io.StringIO()
                code = msg['code']
                with set_resource_mode('cdn'):
                    try:
                        out = exec_with_return(code, stdout=stdout, stderr=stderr)
                    except Exception:
                        out = None
                    mime_type = None
                    if (isinstance(out, (Model, Viewable, Viewer)) or
                        any(pane.applies(out) for pane in CONVERT_PANE_TYPES)):
                        try:
                            _, content = _model_json(as_panel(out), msg['target'])
                            mime_type = 'application/bokeh'
                        except Exception:
                            content = None
                            warnings.warn(
                                f'Could not render {out!r} generated from executed code directive:\n\n{code}\n\n'
                                f'Failed with following error:\n\n{tb.format_exc()}'
                            )
                    elif out is not None:
                        try:
                            content, mime_type = format_mime(out)
                        except Exception:
                            content = None
                            warnings.warn(f'Could not render {out!r} generated from executed code directive: {code}')
                    else:
                        content = None
                    js, js_exports, js_modules, css, global_exports = extract_extensions(code)
                pipe.send((content, mime_type, stdout.getvalue(), stderr.getvalue(), js, js_exports, js_modules, css, global_exports))
            cur_task = asyncio.current_task()
            await asyncio.gather(*(t for t in asyncio.all_tasks() if t is not cur_task), return_exceptions=True)
        asyncio.get_event_loop().run_until_complete(loop())
        pipe.close()

    @classmethod
    def terminate(cls, *args):
        """
        Terminates a running process.
        """
        for source in cls._exec_state:
            cls._kill(source)

    @classmethod
    def _launch_process(cls, source, timeout=5):
        """
        Launches a process to execute code in.
        """
        cls._exec_state[source]['conn'], child_conn = Pipe()
        cls._exec_state[source]['process'] = process = get_context('spawn').Process(
            target=cls._execution_process, args=(child_conn,), daemon=True
        )
        process.start()

    @classmethod
    def _kill(cls, source, clear=True):
        state = cls._exec_state[source]
        if state['process']:
            try:
                state['conn'].send({'type': 'close'})
                state['process'].join()
            except Exception:
                state['process'].terminate()
        if clear:
            del cls._exec_state[source]

    def run(self):
        outdir = self.state.document.settings.env.app.outdir
        cache_path = pathlib.Path(str(outdir)) / '.pyodide'
        cache_path.mkdir(exist_ok=True)
        current_source = self.state_machine.get_source()
        if current_source not in self._exec_state:
            with open(current_source, encoding='utf-8') as src_doc:
                content = src_doc.read()
            content_hash = hashlib.md5(content.encode('utf-8')).hexdigest()
            cache_file = cache_path / f'{content_hash}.json'
            if cache_file.is_file():
                with open(cache_file, encoding='utf-8') as cf:
                    self._exec_state[current_source]['cache'] = json.load(cf)
            else:
                self._launch_process(current_source)
            total = content.count('{pyodide}')
            self._exec_state[current_source]['total'] = total
            self._exec_state[current_source]['hash'] = content_hash
        else:
            self._exec_state[current_source]['count'] += 1
        state = self._exec_state[current_source]
        count = state['count']
        cache_file = cache_path / f'{state["hash"]}.json'

        # Construct code block node
        classes = 'pyodide'
        if 'class' in self.options:
            classes += f" {self.options['class']}"
        self.options['class'] = [classes]
        self.options['id'] = cellid = f'codecell{count}-py'
        roles.set_classes(self.options)
        code = '\n'.join(self.content)
        doctree_node = nodes.literal_block(code, code, **self.options)
        doctree_node['language'] = 'python'

        if self.options.get('skip-embed'):
            return [doctree_node]

        # Send execution request to client and wait for result
        conn = state['conn']
        try:
            code_hash = hashlib.md5(code.encode('utf-8')).hexdigest()
            if code_hash in state['cache']:
                result = state['cache'][code_hash]
            elif conn:
                conn.send({'type': 'execute', 'target': f'output-{cellid}', 'code': code})
                if conn.poll(60):
                    state['cache'][code_hash] = result = conn.recv()
                else:
                    raise RuntimeError('Timed out')
            else:
                raise RuntimeError('Process was shut down')
        except Exception:
            self._kill(current_source, clear=False)
            return [doctree_node]
        finally:
            # If we are in the last pyodide cell kill the process
            if (count+1) == state['total']:
                self._kill(current_source)
                cache_file.write_text(json.dumps(state['cache']))

        output, mime_type, stdout, stderr, js, js_exports, js_modules, css, global_exports = result

        # Write out resources
        resources = {
            'css': css,
            'js': js,
            'js_exports': js_exports,
            'js_modules': js_modules
        }
        write_resources(outdir, current_source, resources)

        stdout_style = 'style="display: block;"' if stdout else ''
        stdout_html = f'<pre id="stdout-{cellid}" class="pyodide-stdout" {stdout_style}>{escape(stdout)}</pre>'
        stdout_node = nodes.raw(stdout, stdout_html, format='html')

        stderr_style = 'style="display: block;"' if stderr else ''
        stderr_html = f'<pre id="stderr-{cellid}" class="pyodide-stderr" {stderr_style}>{escape(stderr)}</pre>'
        stderr_node = nodes.raw(stderr, stderr_html, format='html')

        rendered_nodes = [doctree_node, stdout_node, stderr_node]
        if output is None:
            return rendered_nodes

        # Ensure we wait for all JS module exports to be initialized
        exports = ' && '.join(f'window.{export}' for export in global_exports)
        if exports:
            exports = f' && {exports}'

        script = ""
        if mime_type == 'text/plain':
            output = f'<pre>{output}</pre>'
        elif mime_type == 'application/bokeh':
            script = f"""
            <script>
              async function embed_bokeh_{count} () {{
                if (window.Bokeh && window.Bokeh.Panel{exports}) {{
                  await Bokeh.embed.embed_item({output})
                }} else {{
                   setTimeout(embed_bokeh_{count}, 200)
                }}
              }};
              embed_bokeh_{count}()
            </script>
            """
            output = ""
        else:
            output = f'<div class="pyodide-output-wrapper">{output}</div>'
        output_html = f'<div id="output-{cellid}" class="pyodide-output embedded">{output}</div>\n{script}'
        output_node = nodes.raw(output, output_html, format='html')
        return rendered_nodes+[output_node]

def write_worker(app: Sphinx, exc):
    if app.builder.format != 'html' or exc:
        return
    pyodide_conf = app.config.nbsite_pyodide_conf
    builddir = Path(app.builder.outdir)
    staticdir = builddir / '_static'

    # Render Web Worker
    web_worker = WEB_WORKER_TEMPLATE.render({
        'PYODIDE_URL': pyodide_conf['PYODIDE_URL'],
        'env_spec': ', '.join([repr(req) for req in pyodide_conf['requirements']]),
        'setup_code': pyodide_conf['setup_code'],
        'autodetect_deps': pyodide_conf['autodetect_deps'],
        'requires': json.dumps(pyodide_conf['requires'])
    })
    with open(staticdir/ 'PyodideWebWorker.js', 'w', encoding='utf-8') as f:
        f.write(web_worker)
    worker_setup = WORKER_HANDLER_TEMPLATE.render(
        scripts=pyodide_conf['scripts']
    )
    with open(staticdir/ 'WorkerHandler.js', 'w', encoding='utf-8') as f:
        f.write(worker_setup)

    if not pyodide_conf['enable_pwa']:
        return

    # Render service worker
    service_worker = SERVICE_WORKER_TEMPLATE.render({
        'project': app.config.project,
        'version': app.config.version,
        'pre_cache': ', '.join([repr(req) for req in pyodide_conf['precache']]),
        'cache_patterns': ', '.join([repr(req) for req in pyodide_conf['cache_patterns']])
    })
    with open(builddir / 'PyodideServiceWorker.js', 'w', encoding='utf-8') as f:
        f.write(service_worker)
    service_handler = SERVICE_HANDLER_TEMPLATE.render()
    with open(staticdir/ 'ServiceHandler.js', 'w', encoding='utf-8') as f:
        f.write(service_handler)

    # Render manifest
    site_manifest = WEB_MANIFEST_TEMPLATE.render({
        'name': app.config.html_title,
    })
    with open(builddir / 'site.webmanifest', 'w', encoding='utf-8') as f:
        f.write(site_manifest)


def init_conf(app: Sphinx) -> None:
    pyodide_conf = dict(DEFAULT_PYODIDE_CONF, **app.config.nbsite_pyodide_conf)

    out_dir = pathlib.Path(str(app.outdir))
    resource_file = out_dir / RESOURCE_FILE
    if resource_file.is_file():
        resource_file.unlink()

    app.config.nbsite_pyodide_conf = pyodide_conf
    app.config.html_static_path.append(
        str((HERE /'_static' ).absolute())
    )
    if pyodide_conf['enable_pwa']:
        app.add_js_file('ServiceHandler.js')
    for script in pyodide_conf['scripts']:
        app.add_js_file(script)

def html_page_context(
    app: Sphinx,
    pagename: str,
    templatename: str,
    context: Dict[str, Any],
    doctree: nodes.document,
) -> None:
    if not doctree:
        return

    if app.config.nbsite_pyodide_conf['enable_pwa']:
        relpath = '/'.join(['..']*pagename.count('/'))
        if relpath:
            relpath += '/'
        context[
            "metatags"
        ] += f"""
            <link rel="manifest" href="{relpath}site.webmanifest"/>
            """

    # Add additional resources extracted from pn.extension calls
    sourcename = context['sourcename'].replace('.txt', '')

    resource_file = pathlib.Path(str(app.outdir)) / RESOURCE_FILE
    if resource_file.is_file():
        resources = json.loads(resource_file.read_text())
    else:
        resources = {}
    code_resources = [
        (r['js'], r['js_exports'], r['js_modules'], r['css'])
        for filename, r in resources.items()
        if filename.endswith(sourcename)
    ]
    if code_resources:
        extra_js, js_exports, js_modules, extra_css = code_resources[0]
    else:
        extra_js, js_exports, js_modules, extra_css = [], {}, {}, []
    extra_css += app.config.nbsite_pyodide_conf.get('extra_css', [])
    context["script_files"] += extra_js
    context["css_files"] += extra_css

    module_tags = app.config.nbsite_pyodide_conf['preamble']
    for export, module in js_modules.items():
        module_tags += JS_MODULE_TAG.format(file=module)
    for export, module in js_exports.items():
        module_tags += JS_MODULE_EXPORT.format(name=export, file=module)
    if module_tags:
        context["body"] = f'{module_tags}{context["body"]}'

    # Remove Scripts and CSS from page if no pyodide cells are found.
    if any(
        'pyodide' in cb.attributes.get('classes', [])
        for cb in doctree.traverse(nodes.literal_block)
    ):
        return

    # Remove JS files
    pyodide_scripts = (
        app.config.nbsite_pyodide_conf['scripts'] +
        ['_static/run_cell.js', '_static/WorkerHandler.js'] +
        app.config.html_js_files
    )

    context["script_files"] = [
        ii for ii in context["script_files"]
        if ii not in pyodide_scripts
    ]

    # Remove pyodide CSS files
    context["css_files"] = [
        ii for ii in context["css_files"]
        if ii not in ['_static/runbutton.css']
    ]


def setup(app):
    """Setup sphinx-gallery sphinx extension"""
    app.add_config_value('nbsite_pyodide_conf', DEFAULT_PYODIDE_CONF, 'html')

    app.connect('builder-inited', init_conf)
    app.connect('build-finished', write_worker)
    app.connect('html-page-context', html_page_context)
    app.connect('build-finished', PyodideDirective.terminate)

    app.add_directive('pyodide', PyodideDirective)

    app.add_css_file('runbutton.css')
    app.add_js_file('run_cell.js')
    app.add_js_file('WorkerHandler.js')

    return {
        'version': '0.5',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import glob

from setuptools import find_packages, setup

NAME = 'nbsite'
DESCRIPTION = 'Build a tested, sphinx-based website from notebooks.'

# duplicated from pyct.build.get_setup_version until pyct[build] >0.4.6 lands
def get_setup_version(root, reponame):
    """
    Helper to get the current version from either git describe or the
    .version file (if available) - allows for param to not be available.
    Normally used in setup.py as follows:
    >>> from pyct.build import get_setup_version
    >>> version = get_setup_version(__file__, reponame)  # noqa
    """
    import json
    import os

    filepath = os.path.abspath(os.path.dirname(root))
    version_file_path = os.path.join(filepath, reponame, '.version')
    try:
        from param import version
    except Exception:
        version = None
    if version is not None:
        return version.Version.setup_version(filepath, reponame, archive_commit="$Format:%h$")
    print(
        "WARNING: param>=1.6.0 unavailable. If you are installing a package, "
        "this warning can safely be ignored. If you are creating a package or "
        "otherwise operating in a git repository, you should install param>=1.6.0."
    )
    return json.load(open(version_file_path, 'r'))['version_string']

setup_args = dict(
    name=NAME,
    version=get_setup_version(__file__, NAME),
    description=DESCRIPTION,
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author='HoloViz developers',
    author_email='developers@holoviz.org',
    maintainer='HoloViz developers',
    maintainer_email='developers@holoviz.org',
    url='https://nbsite.holoviz.org',
    project_urls={
        'Documentation': 'https://nbsite.holoviz.org/',
        'Source Code': 'https://github.com/holoviz-dev/nbsite',
        'Bug Tracker': 'https://github.com/holoviz-dev/nbsite/issues'
    },
    packages=find_packages(),
    python_requires='>=3.9',
    install_requires=[
        'param >=1.7.0',
        'pyviz_comms',
        'ipykernel',
        'nbformat',
        'nbconvert',
        'jupyter_client',
        'myst-nb >=1.1',
        'sphinx-design',
        'notebook',
        'sphinx >=7',
        'beautifulsoup4',
        'jinja2',
        'pillow',
        'portalocker',
        'pydata-sphinx-theme >=0.15,<0.16',
        'myst-parser >=3',
        'sphinx-copybutton',
        'sphinx-design',
        'sphinxext-rediraffe',
    ],
    extras_require= {
        'refman':[
            'graphviz',
        ],
        'gallery':[
            'selenium',
            'phantomjs'
        ],
        'tests':[
            'flake8',
            'pytest >=3.9.1',
            'pre-commit',
        ],
        'build': [
            "setuptools",
            "param >=1.6.1",
        ]
    },
    include_package_data=True,
    license='BSD-3',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
    ],
    # TODO: will be transferring to nbsite command...
    scripts = glob.glob("scripts/*.py"),
    entry_points={
        'console_scripts': [
            'nbsite = nbsite.__main__:main'
        ]
    },
)

if __name__=="__main__":
    setup(**setup_args)



import warnings

warnings.filterwarnings("ignore")

import matplotlib as mpl

mpl.use('agg')

try:
    import holoviews.plotting.bokeh  # noqa
except Exception:
    pass

try:
    import holoviews.plotting.mpl as hmpl
    hmpl.MPLPlot.fig_alpha = 0
except Exception:
    pass
